################ Copyright 2005-2018 Team GoldenEye: Source #################
#
# This file is part of GoldenEye: Source's Python Library.
#
# GoldenEye: Source's Python Library is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or(at your option) any later version.
#
# GoldenEye: Source's Python Library is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GoldenEye: Source's Python Library.
# If not, see <http://www.gnu.org/licenses/>.
#############################################################################
from . import GEScenario
from .Utils import GetPlayers, _
from .Utils.GEPlayerTracker import GEPlayerTracker
from GEUtil import Color
import GEUtil, GEMPGameRules as GERules, GEGlobal as Glb, GEWeapon

import math

USING_API = Glb.API_VERSION_1_2_0

TR_ELIMINATED = "eliminated" # If we're eliminated and can not join the round.  Starts true but is set to false upon successful spawn.  Used to track who can spawn.
TR_WASINPLAY = "wasinplay" # If we played in this round.  Starts false but is set to true for all players in the round when it inits.  Used for progress bars.
TR_MARKED = "marked"
TR_LASTKILLTIME = "lastkill"

CLR_MI6_BAR = Color( 100, 184, 234, 220 )
CLR_JANUS_BAR = Color( 206, 43, 43, 255 )
CLR_DM_BAR = Color( 170, 170, 170, 220 )
CLR_SHOWDOWN_BAR = Color( 180, 225, 255, 170 )

# Just a heavily modified version of YOLT.
class OBIEX( GEScenario ):
    def __init__( self ):
        GEScenario.__init__( self )

        self.pltracker = GEPlayerTracker( self )

        self.game_foes_orig = 0
        self.game_MI6_foes_orig = 0
        self.game_Janus_foes_orig = 0
        self.game_foes = None

        self.nextchecktime = 0

    def GetTeamPlay( self ):
        return Glb.TEAMPLAY_NONE

    def GetPrintName( self ):
        return "One Bullet is Enough X"

    def GetScenarioHelp( self, help_obj ):
        help_obj.SetDescription( "You are given a knife and a Golden Gun with only one bullet.\n\nEvery hit on your opponent is an instant kill, and will reward you with a bullet.\n\nYou are limited to two lives per round. The last man standing will be given a point for every player who was eliminated that round!\n\nTeamplay: Never\n\nV1" )

    def GetGameDescription( self ):
        if GERules.IsTeamplay():
            return "Team OBIE X"
        else:
            return "OBIE X"

    def OnLoadGamePlay( self ):
        GERules.DisableSuperfluousAreas()
        self.CreateCVar( "yolt_timeperplayer", "30", "How many seconds are added to the clock for each player in the round." )        
        self.CreateCVar( "yolt_basetime", "0", "How many seconds are added to the clock regardless of the playercount." )        

    def OnUnloadGamePlay( self ):
        super( OBIEX, self ).OnUnloadGamePlay()
        self.pltracker = None

    def OnPlayerConnect( self, player ):
        self.pltracker[player][TR_ELIMINATED] = True
        self.pltracker[player][TR_WASINPLAY] = False
        self.pltracker[player][TR_MARKED] = False
        self.pltracker[player][TR_LASTKILLTIME] = GEUtil.GetTime()

    def OnPlayerDisconnect( self, player ):
        if GERules.IsRoundLocked() and self.__IsInPlay( player ):
            self.__DecreaseFoes( player )

    def CanPlayerChangeTeam( self, player, oldTeam, newTeam, wasforced ):
        if wasforced: # Make sure we can't eliminate people with autobalance.  Seriously, without this check you can eliminate an entire team.
            return True
        
        if GERules.IsRoundLocked():
            if oldTeam == Glb.TEAM_SPECTATOR: # Switching from spectator.  But we're only allowed to join the round if it isn't locked.
                GEUtil.PopupMessage( player, "#GES_GPH_CANTJOIN_TITLE", "#GES_GPH_CANTJOIN" )
                self.pltracker[player][TR_ELIMINATED] = True  
            elif not self.pltracker[player][TR_ELIMINATED]: # We switched during a locked round, eliminate us.
                # Tell everyone who was eliminated
                msg = _( "#GES_GP_YOLT_ELIMINATED", player.GetCleanPlayerName() )
                GEUtil.PostDeathMessage( msg )
                # Tell plugins who was eliminated
                GEUtil.EmitGameplayEvent( "obiex_eliminated", str( player.GetUserID() ), str(-1), str(-1) )
                # Tell the victim they are eliminated
                GEUtil.PopupMessage( player, "#GES_GPH_ELIMINATED_TITLE", "#GES_GPH_ELIMINATED" )
                # Decrease the foes, which will call initfoes if we're the first one out.
                self.__DecreaseFoes( player )
        else:
            player.ChangeTeam( newTeam, True ) # Force team change instead, so we don't suicide and start the round.
            return False # Since we force changed team we don't want to finish the normal team swap.

        return True

    def OnPlayerSpawn( self, player ):
        # Update our tracker variables
        self.pltracker[player][TR_ELIMINATED] = False
        self.pltracker[player][TR_LASTKILLTIME] = GEUtil.GetTime()

        self.__CheckObjStatus( player, 45 )

        player.StripAllWeapons()

        player.GiveNamedWeapon( "weapon_knife", 0 )
        player.GiveNamedWeapon( "weapon_golden_gun", 1, True )
        player.WeaponSwitch( "weapon_golden_gun" )

        if not GERules.IsRoundLocked():
            self.__InitOpenroundBars()

        if player.IsInitialSpawn():
            GEUtil.PopupMessage( player, "One Bullet is Enough", "You have one bullet. Use it wisely.  Knife is an instant kill." )

    def OnPlayerObserver( self, player ):
        # Init the tracking bars
        self.__InitObserverBars( player )

    def OnRoundBegin( self ):
        super( OBIEX, self ).OnRoundBegin()

        # Remove latent bars and let everyone know it's an open round
        self.__InitOpenroundBars()

        GERules.UnlockRound()
        GERules.DisableWeaponSpawns()
        GERules.DisableAmmoSpawns()
        GERules.DisableArmorSpawns()

        GERules.UnlockRound()
        GERules.GetRadar().SetForceRadar( False )

        self.pltracker.SetValueAll( TR_WASINPLAY, False )
        self.pltracker.SetValueAll( TR_MARKED, False )
        self.pltracker.SetValueAll( TR_LASTKILLTIME, GEUtil.GetTime() )

        self.game_foes_orig = 0
        self.game_MI6_foes_orig = 0
        self.game_Janus_foes_orig = 0
        self.game_foes = None

        # Start dynamic timer
        self.__UpdateRoundTimer( 0 )

        GERules.SetSpawnInvulnTime( 0, True )

    def OnPlayerKilled( self, victim, killer, weapon ):
        # Execute default DM scoring behavior
        killerStartScore = 0

        if killer:
            killerStartScore = killer.GetRoundScore()

        GEScenario.OnPlayerKilled( self, victim, killer, weapon )

        if GERules.GetNumInRoundPlayers() < 2:
            if not victim: # We don't do YOLT logic if there aren't enough players for it.
                return

            victim.StripAllWeapons() # This is a super fringe exploit but it's possible to kill yourself before anyone joins and drop a weapon to pick up later.
            return

        if killer and killer.GetRoundScore() - killerStartScore == 1: # We gained a point so this must have been a good kill!
            killer.GiveAmmo( "goldenammo", 1 )
            self.pltracker[killer][TR_LASTKILLTIME] = GEUtil.GetTime()

        # Lock the round (if not done already)
        GERules.LockRound()
        if self.game_foes_orig == 0: # If we don't have any foes we'll need to init them.
            self.__InitFoes()

        if victim.GetDeaths() >= 2 and not self.pltracker[victim][TR_ELIMINATED]: # Make sure we're not already eliminated since switching teams can do that.
            # Tell everyone who was eliminated
            msg = _("#GES_GP_YOLT_ELIMINATED", victim.GetCleanPlayerName())
            # msg = _(str(victim.GetCleanPlayerName()))
            GEUtil.PostDeathMessage( msg )
            # Tell plugins who was eliminated
            GEUtil.EmitGameplayEvent( "obiex_eliminated", str( victim.GetUserID() ), str( killer.GetUserID() if killer else -1 ), str(weapon.GetClassname().lower() if weapon else -1), "", True )
            # Tell the victim they are eliminated
            GEUtil.PopupMessage( victim, "#GES_GPH_ELIMINATED_TITLE", "#GES_GPH_ELIMINATED" )
            # Decrease the foes, eliminate the victim
            self.__DecreaseFoes( victim )

        victim.StripAllWeapons() # This prevents the victim from dropping weapons, which might confuse players since there are no pickups.

    def CanPlayerRespawn( self, player ):
        if GERules.IsRoundLocked() and self.pltracker[player][TR_ELIMINATED]:
            player.SetScoreBoardColor( Glb.SB_COLOR_ELIMINATED )
            return False

        player.SetScoreBoardColor( Glb.SB_COLOR_NORMAL )
        return True

    def CanMatchEnd( self ):
        if GERules.IsIntermission() or GERules.GetNumInRoundPlayers() < 2: #We just finished a round or it's not possible to get kills.
            return True
        else:
            return False

    def CalculateCustomDamage( self, victim, info, health, armor ):
        weapon = GEWeapon.ToGEWeapon( info.GetWeapon() )
        if weapon and weapon.GetClassname().lower() != "weapon_golden_gun":
            return 160, 160

        return health, armor

    # Update everyone's timers and see if their score should be adjusted.
    def OnThink(self):
        curtime = GEUtil.GetTime()

        if self.game_foes_orig == 0:
            return # Haven't started the round yet!

        if curtime < self.nextchecktime:
            return # We don't want to do all these calculations more than we need to.

        for player in GetPlayers():
            if not player.IsDead():
                plCurrentSeconds = max(int( self.pltracker[player][TR_LASTKILLTIME] + 45 - curtime ), 0)

                # Update our displays and check to see if we should be an objective.
                self.__CheckObjStatus(player, plCurrentSeconds)
                self.__UpdateStealthTimer(player, plCurrentSeconds)

    def __UpdateRoundTimer( self, playercount, onlydecrease = False ):
        playertime = int( GEUtil.GetCVarValue( "yolt_timeperplayer" ) )
        basetime = int( GEUtil.GetCVarValue( "yolt_basetime" ) )

        #Dynamic timer disabled.
        if playertime == 0 and basetime == 0:
            return

        #Setting unlimited time, which we have to do here instead of directly so we don't mess with static roundtimes.
        if playercount == 0:
            GERules.SetRoundTimeLeft(0)
            GERules.AllowRoundTimer( False ) # We do both of these so it hides the match timer and flashes green when the time is started.
            return

        GERules.AllowRoundTimer( True ) # Reenable it now that the timer is started again.
        newtime = basetime + playercount * playertime

        #Make sure we're decreasing and not increasing the roundtime if we want to avoid doing that.
        if onlydecrease:
            timeleft = int(GERules.GetRoundTimeLeft())

            if (timeleft > newtime):
                GERules.SetRoundTimeLeft(newtime)
        else:
            GERules.SetRoundTimeLeft(newtime)

    def __UpdateStealthTimer( self, player, seconds ):
        if self.__IsShowdown():
            GEUtil.HudMessage( player, "Showdown!  Use the radar to find your opponent.", -1, 0.745, GEUtil.Color( 150, 150, 150, 255 ), 1.5, 0 )
            return


        if seconds == 0:
            GEUtil.HudMessage( player, "Your position is revealed!  Get a kill to regain stealth!", -1, 0.745, GEUtil.Color( 255, 0, 50, 255 ), 1.5, 0 )
        else:
            GEUtil.HudMessage( player, "Stealth: " + str(self.__SecondsToClockTime(seconds)), -1, 0.745, GEUtil.Color( 150, 150, 150, 255 ), 1.5, 0 )

    # Check to see if we should become an objective for other players
    def __CheckObjStatus( self, player, seconds ):
        if seconds == 0 and not self.__IsShowdown() and not self.pltracker[player][TR_MARKED]:
            radar = GERules.GetRadar()
            radar.SetupObjective( player, Glb.TEAM_NONE, "", "Revealed Player", GEUtil.Color( 208, 179, 64, 255 ), 0 ) # Minimum distance of 0 because this is supposed to be punishing!
            radar.AddRadarContact( player, Glb.RADAR_TYPE_PLAYER, True, "", GEUtil.Color( 208, 179, 64, 255 ) )
            self.pltracker[player][TR_MARKED] = True
        elif ( seconds > 0 or self.__IsShowdown() ) and self.pltracker[player][TR_MARKED]:
            radar = GERules.GetRadar()
            radar.DropRadarContact( player )
            self.pltracker[player][TR_MARKED] = False

    # Convert an integer value to a string that can be displayed on the HUD
    def __SecondsToClockTime( self, seconds ):
        secondcount = seconds % 60
        secondstring = str(secondcount)
        
        if secondcount <= 9:
            secondstring = "0" + secondstring

        # Time to be displayed is minutes:seconds
        time = str(int(math.floor(seconds/60))) + ":" + secondstring
        return time

    def __InitFoes( self ):
        if self.game_foes_orig != 0:
            # We are already init'd
            return
        
        if GERules.IsTeamplay():
            mi6_count = GERules.GetNumInRoundTeamPlayers( Glb.TEAM_MI6 )
            janus_count = GERules.GetNumInRoundTeamPlayers( Glb.TEAM_JANUS )
            self.game_foes_orig = mi6_count + janus_count
            self.game_MI6_foes_orig = mi6_count
            self.game_Janus_foes_orig = janus_count
            self.game_foes = [mi6_count, janus_count]
            playercount = min(mi6_count, janus_count) * 2
            # Display the foes progress bars (note color and values are swapped since these are "foes")
            GEUtil.InitHudProgressBar( Glb.TEAM_MI6_NO_OBS, 0, "#GES_GP_FOES", Glb.HUDPB_SHOWVALUE, janus_count, -1, 0.02, 0, 10, CLR_JANUS_BAR, janus_count )
            GEUtil.InitHudProgressBar( Glb.TEAM_JANUS_NO_OBS, 0, "#GES_GP_FOES", Glb.HUDPB_SHOWVALUE, mi6_count, -1, 0.02, 0, 10, CLR_MI6_BAR, mi6_count )
        else:
            # DM foes is just a number
            self.game_foes_orig = GERules.GetNumInRoundPlayers()
            playercount = self.game_foes_orig
            # Subtract one because we don't count the local player as a "foe"
            GEUtil.InitHudProgressBar( Glb.TEAM_NO_OBS, 0, "#GES_GP_FOES", Glb.HUDPB_SHOWVALUE, self.game_foes_orig - 1, -1, 0.02, 0, 10, CLR_DM_BAR, self.game_foes_orig - 1 )
            # Copy our origin into the tracker
            self.game_foes = self.game_foes_orig

        for player in GetPlayers():
            if not self.pltracker[player][TR_ELIMINATED]:
                self.pltracker[player][TR_WASINPLAY] = True

        self.__UpdateRoundTimer( playercount )

        # Start everyone off on the same footing.
        self.pltracker.SetValueAll( TR_LASTKILLTIME, GEUtil.GetTime() )
        
        # Make sure we don't need to force the radar.
        self.__CheckWin()

        # Initialize observer bars
        self.__InitObserverBars( Glb.TEAM_OBS )
        
    def __DecreaseFoes( self, victim ):
        # Init the foes if needed
        if self.game_foes_orig == 0:
            self.__InitFoes()

        # Mark the victim as eliminated
        self.pltracker[victim][TR_ELIMINATED] = True
        victim.SetScoreBoardColor( Glb.SB_COLOR_ELIMINATED ) # Make sure they get the color change if they're the last to die.

        playercount = 1

        if GERules.IsTeamplay():
            # Team foes is ( mi6, janus )
            if victim.GetTeamNumber() == Glb.TEAM_MI6:
                self.game_foes[0] -= 1
            else:
                self.game_foes[1] -= 1

            # Update non-observers (note the reversal of scores)
            GEUtil.UpdateHudProgressBar( Glb.TEAM_MI6_NO_OBS, 0, self.game_foes[1] )
            GEUtil.UpdateHudProgressBar( Glb.TEAM_JANUS_NO_OBS, 0, self.game_foes[0] )

            # Update observers
            GEUtil.UpdateHudProgressBar( Glb.TEAM_OBS, 0, self.game_foes[0] )
            GEUtil.UpdateHudProgressBar( Glb.TEAM_OBS, 1, self.game_foes[1] )

            # Update timer max
            playercount = min(self.game_foes[0], self.game_foes[1]) * 2
        else:
            # DM foes is just a number
            self.game_foes -= 1
            # We subtract 1 here to account for the local player (prevents "1 / X" at win)
            GEUtil.UpdateHudProgressBar( Glb.TEAM_NO_OBS, 0, self.game_foes - 1 )
            # Update observers, we don't subtract 1 anymore
            GEUtil.UpdateHudProgressBar( Glb.TEAM_OBS, 0, self.game_foes )

            # Update timer max
            playercount = self.game_foes

            #We do this here so the victim doesn't see themselves taken off the foe meter.
            self.__InitObserverBars( victim )

        # Check if we have a winner
        self.__CheckWin()

        self.__UpdateRoundTimer(playercount, True)                                       

    def __CheckWin( self ):
        if not self.game_foes:
            return

        if GERules.IsTeamplay():
            # Default no winner
            winner = None
            winnercount = 1
            if self.game_foes[0] <= 0:
                # MI6 has no foes, Janus won
                winner = GERules.GetTeam( Glb.TEAM_JANUS )
                winnercount = self.game_foes[1]
                losercount = self.game_MI6_foes_orig
            elif self.game_foes[1] <= 0:
                # Janus has no foes, MI6 won
                winner = GERules.GetTeam( Glb.TEAM_MI6 )
                winnercount = self.game_foes[0]
                losercount = self.game_Janus_foes_orig
            elif self.__IsShowdown():
                # When we are down to 1 player on each team, force show the radar
                GERules.GetRadar().SetForceRadar( True )
                self.__InitShowdownBars()

            if winner:
                # Split up the spoils among the surviving players on the winning team.
                for player in GetPlayers():
                    if self.__IsInPlay( player ):
                        player.AddRoundScore( int(round(losercount/winnercount)) )
                
                winner.AddRoundScore( losercount ) # The team gets all of it, though.
                GERules.EndRound()
        else:
            # DM uses a number
            if self.game_foes <= 1:
                # There is only 1 player left, find who that is
                for player in GetPlayers():
                    if self.__IsInPlay( player ):
                        player.AddRoundScore( self.game_foes_orig ) #They still might not win, but if they don't manage to win with this many points they don't deserve it.
                        GEUtil.EmitGameplayEvent( "obiex_lastmanstanding", str( player.GetUserID() ), str(self.game_foes_orig), "", "", True )
                        break
                # Round is over
                GERules.EndRound()
            elif self.__IsShowdown():
                GERules.GetRadar().SetForceRadar( True )
                self.__InitShowdownBars()

    def __InitObserverBars( self, player ):
        if not GERules.IsRoundLocked():
            self.__InitOpenroundBars()
            return
        
        # We don't init if there are no foes
        if not self.game_foes:
            return

        if self.__IsShowdown():
            self.__InitShowdownBars()
            return

        # Remove latent bars
        GEUtil.RemoveHudProgressBar( player, 0 )
        GEUtil.RemoveHudProgressBar( player, 1 )

        # Initialize and update
        if GERules.IsTeamplay():
            GEUtil.InitHudProgressBar( player, 0, "##TEAM_MI6: ", Glb.HUDPB_SHOWVALUE, self.game_MI6_foes_orig, 0.35, 0.14, 0, 10, CLR_MI6_BAR, self.game_foes[0] )
            GEUtil.InitHudProgressBar( player, 1, "##TEAM_JANUS: ", Glb.HUDPB_SHOWVALUE, self.game_Janus_foes_orig, 0.5, 0.14, 0, 10, CLR_JANUS_BAR, self.game_foes[1] )
        else:
            foes_orig = self.game_foes_orig
            if self.pltracker.GetValue( player, TR_WASINPLAY, False ):
                # Previously in-game players see a different original foes count
                foes_orig -= 1
                
            
            GEUtil.InitHudProgressBar( player, 0, "#GES_GP_FOES", Glb.HUDPB_SHOWVALUE, foes_orig, -1, 0.14, 0, 10, CLR_DM_BAR, self.game_foes )

    def __InitShowdownBars( self ):
        # Remove latent bars
        GEUtil.RemoveHudProgressBar( None, 0 )
        GEUtil.RemoveHudProgressBar( None, 1 )

        # Prop up nice showdown bars for everyone
        GEUtil.InitHudProgressBar( Glb.TEAM_NO_OBS, 0, "Showdown!", Glb.HUDPB_TITLEONLY, y=0.02, color=CLR_SHOWDOWN_BAR )
        GEUtil.InitHudProgressBar( Glb.TEAM_OBS, 0, "Showdown!", Glb.HUDPB_TITLEONLY, y=0.14, color=CLR_SHOWDOWN_BAR )

    def __InitOpenroundBars( self ):
        # Remove latent bars
        GEUtil.RemoveHudProgressBar( None, 0 )
        GEUtil.RemoveHudProgressBar( None, 1 )

        # Prop up nice showdown bars for everyone
        GEUtil.InitHudProgressBar( Glb.TEAM_NO_OBS, 0, "#GES_GP_OPENROUND", Glb.HUDPB_TITLEONLY, y=0.02, color=CLR_SHOWDOWN_BAR )
        GEUtil.InitHudProgressBar( Glb.TEAM_OBS, 0, "#GES_GP_OPENROUND", Glb.HUDPB_TITLEONLY, y=0.14, color=CLR_SHOWDOWN_BAR )

    def __IsShowdown( self ):
        return (self.game_foes[0] == 1 and self.game_foes[1] == 1) if GERules.IsTeamplay() else self.game_foes == 2

    def __IsInPlay( self, player ):
        try:
            return player.GetTeamNumber() != Glb.TEAM_SPECTATOR 	\
                    and not self.pltracker[player][TR_ELIMINATED]
        except KeyError:
            pass
